/* Atomix -- a little mind game about atoms and molecules.
 * Copyright (C) 1999 Jens Finke
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <sys/types.h>
#include <dirent.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>

#include <gnome.h>
#include "global.h"
#include "interface.h"
#include "support.h"
#include "board.h"
#include "playfield.h"
#include "main.h"
#include "goal.h"
#include "level.h"
#include "gtk-time-limit.h"

/****************************
 *    global variables
 */
static GtkWidget *app;       /* main application */
static Level *level = NULL;  /* actual level */
static gdouble score = 0.0;  /* players score */
static Theme *theme = NULL;  /* actual used theme */
static Preferences *prefs = NULL; /* game preferences */

enum {
	GAME_NOT_RUNNING,
	GAME_RUNNING,
	GAME_PAUSED
};

gint game_state = GAME_NOT_RUNNING;


/**
 *  dialog callbacks
 */
gint next_level_dlg_cb(gpointer data);
void init_preferences(void);
void save_score(gdouble score);
void game_clean_up(void);

void set_appbar_temporary(char *txt);
int clear_appbar(void *id);
/***************************
 *    functions
 */
GtkWidget* get_app() 
{
	return app;
}

Level* get_actual_level()
{
	return level;
}

Theme* get_actual_theme()
{
	return theme;
}

Preferences* get_preferences()
{
	return prefs;
}

void game_set_score(gdouble value)
{
	GtkWidget *en_score;
	gchar *str_score;

	if(prefs->score_enabled)
	{
		/* set score */
		score = value;
		str_score = g_malloc(10*sizeof(gchar));
		en_score = GTK_WIDGET(lookup_widget(app, "en_score"));
		g_snprintf(str_score, 10, "%.2f", score);
		gtk_label_set_text(GTK_LABEL(en_score), str_score);
	}
}


void game_init()
{
	GtkWidget* clock;

	board_init();
    
	/* connect the timeout function */
	clock = lookup_widget(app, "clock");
	gtk_signal_connect(GTK_OBJECT(clock), "timeout",
			   GTK_SIGNAL_FUNC(game_level_timeout),NULL);
}

void game_new()
{
	LoadResult result;

	game_state = GAME_RUNNING;
	score = 0.0;

	level = level_load_xml(level_get_first_level()/*, &result*/);

	if(level) 
	{
		GtkWidget* dlg;
		board_hide_message(BOARD_MSG_NEW_GAME);

		dlg = create_dlg_first_level();
		gtk_widget_show(dlg);
		gnome_dialog_run(GNOME_DIALOG(dlg));

		game_load_level();    
	}
#ifdef DEBUG
	else
	{
		g_print("game_new(): Level == NULL\n");
	}
#endif
}

void game_undo_move()
{
	if(board_undo_move())
	{
		set_appbar_temporary(_("Undo last move. This costs you 5% of your score."));
		game_set_score(score*0.95);
	}
	else
	{
		set_appbar_temporary(_("No more undo levels."));
	}
}

void game_load_level()
{
	LoadResult result;
	GtkWidget* clock;
	GtkWidget* en_level;
    
	/* load theme if necessary */
	if(theme)
	{
		if(g_strcasecmp(theme->name, level->theme_name) != 0)
		{
			/* the theme changed for this level */
			theme = theme_load_xml(level->theme_name/*, &result*/);
		}
	}    
	else
	{
		theme = theme_load_xml(level->theme_name/*, &result*/);    
	}

	/* init the level */
	board_init_level(level);
    
	/* set levelname */
	en_level = GTK_WIDGET(lookup_widget(app, "en_level"));
	gtk_label_set_text(GTK_LABEL(en_level), level->name);

	/* set score */
	game_set_score(score);

	/* set time */
	clock = GTK_WIDGET(lookup_widget(app, "clock"));
	gtk_clock_set_seconds(GTK_CLOCK(clock), level->time);
	gtk_time_limit_start(GTK_TIME_LIMIT(clock));
}


void game_reload_level()
{
	game_load_level();
}

void game_level_timeout(GtkWidget *widget, gpointer data)
{
	GtkWidget *dlg;
	GtkWidget* clock;
	gint button_nr;

	/* stop the clock */
	clock = GTK_WIDGET(lookup_widget(app, "clock"));
	gtk_time_limit_stop(GTK_TIME_LIMIT(clock));
	gtk_clock_set_seconds(GTK_CLOCK(clock), 0);

	/* show the cursor */
	board_show_normal_cursor();

	score = score * 0.9;
	game_set_score(score);

	dlg = create_dlg_timeout();
	gtk_widget_show(GTK_WIDGET(dlg));
	button_nr = gnome_dialog_run(GNOME_DIALOG(dlg));
	gtk_widget_destroy(GTK_WIDGET(dlg));

	if(button_nr == 0)
	{
		/* player want to quit */
		game_clean_up();
		save_score(score);
		board_view_message(BOARD_MSG_NEW_GAME);
	}
	else
	{	
		/* restart level */
		game_reload_level();
	}
}

void game_level_finished(void)
{
	GtkWidget* clock;
	LoadResult result;
	gchar* next;
	gint secs;
	GtkWidget* dlg = NULL;
	NextLevelDlgInfo* dlg_info;

	/* stop clock */
	clock = GTK_WIDGET(lookup_widget(app, "clock"));
	gtk_time_limit_stop(GTK_TIME_LIMIT(clock));

	/* show cursur */
	board_show_normal_cursor();

	/* retrieve next level */
	next = g_strdup(level->next);    
    
	if(!level_is_last_level(level))
	{
		if(prefs->score_enabled)
		{
			/* show statistics dialog */     
			secs = gtk_time_limit_get_seconds(GTK_TIME_LIMIT(clock));
			dlg = create_dlg_next_level ();
			dlg_info = g_malloc(sizeof(NextLevelDlgInfo));
			dlg_info->timer_id = -1;
			dlg_info->secs = gtk_time_limit_get_seconds(GTK_TIME_LIMIT(clock));
			dlg_info->score = score;
			dlg_info->lb_secs = lookup_widget(dlg, "time");
			dlg_info->lb_score = lookup_widget(dlg, "score");       
			dlg_info->timer_id = gtk_timeout_add(60, next_level_dlg_cb, 
							     dlg_info);
	    
			gtk_widget_show(GTK_WIDGET(dlg));
			gnome_dialog_run(GNOME_DIALOG(dlg));
			gtk_widget_destroy(GTK_WIDGET(dlg));
	    
			if(dlg_info->timer_id != -1)
			{
				gtk_timeout_remove(dlg_info->timer_id);
			}
			g_free(dlg_info);
    
			/* set new score */
			score = score + secs;
			game_set_score(score);
		}

		/* load the next level */
		level_destroy(level);
		level = level_load_xml(next/*, &result*/);
		g_free(next);
       
		game_load_level();
	}
	else
	{
		/* player reached game end */
		game_finished();
	}
}

void set_appbar_temporary(char *txt)
{
	static gint id = -1;
	GnomeAppBar* appbar = GNOME_APPBAR(lookup_widget(app, "appbar"));
	if(id != -1) gtk_timeout_remove(id);
	gnome_appbar_set_status(appbar, txt);
	id = gtk_timeout_add(2000, clear_appbar, &id);    
}

int
clear_appbar(void *data)
{
	int *id = data;
	GnomeAppBar* appbar = GNOME_APPBAR(lookup_widget(app, "appbar"));
	gnome_appbar_refresh(appbar);
	gtk_timeout_remove(*id);
	*id = -1;

	return 1;
}

gint next_level_dlg_cb(gpointer data)
{
	NextLevelDlgInfo* dlg_info = (NextLevelDlgInfo*) data;
    
	if(dlg_info->secs != 0)
	{
		gint minutes, secs;
		gchar* time_str = g_malloc(6*sizeof(gchar));
		gchar* score_str = g_malloc(10*sizeof(gchar));

		dlg_info->secs--;
		dlg_info->score = dlg_info->score + 1.0;
	
		minutes = (dlg_info->secs)/60;
		secs = (dlg_info->secs)%60;
	
		if(secs < 10)
		{
			g_snprintf(time_str, 6, "%i:0%i", minutes, secs);
		}
		else
		{
			g_snprintf(time_str, 6, "%i:%i", minutes, secs);
		}
		g_snprintf(score_str, 10, "%.2f", dlg_info->score);
	
		gtk_label_set_text(GTK_LABEL(dlg_info->lb_secs), time_str);
		gtk_label_set_text(GTK_LABEL(dlg_info->lb_score), score_str);
	}
	else
	{
		gtk_timeout_remove(dlg_info->timer_id);
		dlg_info->timer_id = -1;
	}    

	return TRUE;
}


void game_finished(void)
{
	GtkWidget* clock;
	gint secs;
	GtkWidget* dlg = NULL;
	NextLevelDlgInfo* dlg_info;
	gint button_nr = -1;

	/* stop clock */
	clock = GTK_WIDGET(lookup_widget(app, "clock"));
	
	if(prefs->score_enabled)
	{

		/* show statistics dialog */     
		secs = gtk_time_limit_get_seconds(GTK_TIME_LIMIT(clock));
		dlg = create_dlg_last_level ();
		dlg_info = g_malloc(sizeof(NextLevelDlgInfo));
		dlg_info->timer_id = -1;
		dlg_info->secs = gtk_time_limit_get_seconds(GTK_TIME_LIMIT(clock));
		dlg_info->score = score;
		dlg_info->lb_secs = lookup_widget(dlg, "time");
		dlg_info->lb_score = lookup_widget(dlg, "score");       
		dlg_info->timer_id = gtk_timeout_add(60, next_level_dlg_cb, dlg_info);
	
		gtk_widget_show(GTK_WIDGET(dlg));
		button_nr = gnome_dialog_run(GNOME_DIALOG(dlg));
		gtk_widget_destroy(GTK_WIDGET(dlg));
    
		if(dlg_info->timer_id != -1)
		{
			gtk_timeout_remove(dlg_info->timer_id);
		}
		g_free(dlg_info);

		/* calculate final score */
		score = score + secs;
		save_score(score);
	}

	game_clean_up();
    
	if(button_nr == 1)
	{
		/* player wants new game */
		game_new();
	}
	else
	{
		/* no new game */
		board_view_message(BOARD_MSG_NEW_GAME);
	}
}

void game_clean_up(void)
{
	GtkWidget *lb_name, *lb_score;
	GtkWidget *clock;

	/* clean up */
	lb_name = GTK_WIDGET(lookup_widget(app, "en_level"));
	gtk_label_set_text(GTK_LABEL(lb_name), "");
    
	lb_score = GTK_WIDGET(lookup_widget(app, "en_score"));
	gtk_label_set_text(GTK_LABEL(lb_score), "");
    
	clock = GTK_WIDGET(lookup_widget(app, "clock"));
	gtk_clock_set_seconds(GTK_CLOCK(clock), 0);

	game_state = GAME_NOT_RUNNING;
	if(level) level_destroy(level);
	level = NULL;
	board_clear();
}

void game_pause(void)
{
	GtkWidget *tl;

	if(game_state == GAME_RUNNING)
	{
		tl = lookup_widget(app, "clock");
		gtk_time_limit_stop(GTK_TIME_LIMIT(tl));
	
		board_hide();
		board_view_message(BOARD_MSG_GAME_PAUSED);
		game_state = GAME_PAUSED;
	}
}


void game_continue(void)
{
	GtkWidget *tl;

	if(game_state == GAME_PAUSED)
	{
		board_hide_message(BOARD_MSG_GAME_PAUSED);
		board_show();
	
		tl = lookup_widget(app, "clock");
		gtk_time_limit_start(GTK_TIME_LIMIT(tl));
	
		game_state = GAME_RUNNING;
	}
}

void save_score(gdouble score)
{
	if(score > 0.0)
	{
		gint highscore_pos;
		highscore_pos = gnome_score_log(score, NULL, TRUE);
		gnome_scores_display("Atomix", "atomix", NULL, highscore_pos);	    
	}
}


void init_preferences() 
{
	if(!prefs)
	{
		prefs = g_malloc(sizeof(Preferences));
		prefs->score_enabled = TRUE;
		prefs->time_enabled = TRUE;
	}
}

void create_user_config_dir(void)
{
	DIR* ds = NULL;
	char* home_dir; 
	char* atomix_dir;
	char* themes_dir;
	char* levels_dir;

	home_dir = g_get_home_dir();
 
	atomix_dir = g_strjoin( "/", home_dir, ".atomix", NULL);
	themes_dir = g_strjoin( "/", atomix_dir, "themes", NULL);
	levels_dir = g_strjoin( "/", atomix_dir, "levels", NULL);

	/* first, check whether .atomix exists */    
	ds = opendir(atomix_dir);
	if(ds==NULL)
	{
		/* try to create one */
		if(mkdir(atomix_dir, 0755 )!=0)
		{
			g_print("An error occured creating your .atomix dir!\n");
			return;
		}
		g_print(".atomix created.\n");
	}
	else
	{
		// g_print(".atomix already exists.\n");
		closedir(ds);
	}

	/* check if .atomix/themes exists */
	ds = opendir(themes_dir);
	if(ds==NULL)
	{
		/* try to create one */
		if(mkdir(themes_dir, 0755)!=0)
		{
			g_print("An error occured creating your .atomix/themes dir!\n");
			return;
		}
		g_print(".atomix/themes created.\n");
	}
	else
	{
		// g_print(".atomix/themes already exists.\n");
		closedir(ds);
	}

	/* check if .atomix/levels exists */
	ds = opendir(levels_dir);
	if(ds==NULL)
	{
		/* try to create one */
		if(mkdir(levels_dir, 0755)!=0)
		{
			g_print("An error occured creating your .atomix/levels dir!\n");
			return;
		}
		g_print(".atomix/levels created.\n");
	}
	else
	{
		// g_print(".atomix/levels already exists.\n");
		closedir(ds);
	}
    
	g_free(atomix_dir);
	g_free(themes_dir);
	g_free(levels_dir);
}

int
main (int argc, char *argv[])
{
	bindtextdomain (PACKAGE, PACKAGE_LOCALE_DIR);
	textdomain (PACKAGE);

	gnome_init ("atomix", VERSION, argc, argv);

	init_preferences();

	prefs->score_enabled = (gnome_score_init("atomix") == 0);

	/* make a few initalisations here */
	create_user_config_dir();
	level_create_hash_table();
	theme_create_hash_table();

	app = create_app ();
	game_init();

	gtk_widget_show (app);

	gtk_main ();
	return 0;
}

