/* Atomix -- a little mind game about atoms and molecules.
 * Copyright (C) 1999 Jens Finke
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <fcntl.h>
#include <parser.h> /* from libxml */
#include <tree.h>   /* from libxml */

#include "theme.h"

/*=================================================================
 
  Declaration of internal functions

  ---------------------------------------------------------------*/
void 
theme_fill_img_array(Theme *theme, xmlNodePtr node, TileType type,
		     guint revision);

void
theme_fill_connection_array(Theme *theme, xmlNodePtr node, guint revision);

gchar*
theme_get_theme_name(gchar *filename);

gchar*
theme_get_theme_name_xml(gchar *filename);

void 
theme_save_image_element_xml(gpointer key, gpointer value, gpointer user_data);

void
theme_print_hash_table(gpointer key, gpointer value, gpointer user_data);

void
theme_destroy_name_table_item(gpointer key, gpointer value, gpointer user_data);

void
theme_destroy_image_table_item(gpointer key, gpointer value, gpointer user_data);

void 
theme_add_tile_image(Theme *theme, TileType type, gint id, ThemeElement *element);

void
theme_add_connection_image(Theme* theme, gint id, ThemeElement *element);

void
theme_destroy_element(ThemeElement *element);

void
theme_add_theme_name(gchar *themename, gchar *directory);

/*=================================================================
 
  Theme creation, initialisation and clean up

  ---------------------------------------------------------------*/
Theme*
theme_new()
{
	Theme* theme;
	theme = g_malloc(sizeof(Theme));

	theme->name = NULL;
	theme->path = NULL;
	theme->obstacles = g_hash_table_new((GHashFunc) g_direct_hash, 
					    (GCompareFunc) g_direct_equal);
	theme->moveables =  g_hash_table_new((GHashFunc) g_direct_hash, 
					     (GCompareFunc) g_direct_equal);
	theme->connections =  g_hash_table_new((GHashFunc) g_direct_hash, 
					       (GCompareFunc) g_direct_equal);
	theme->decor = NULL;
	theme->tile_width = 0;
	theme->tile_height = 0;
	theme->animstep = 4;
	theme->bgcolor = NULL;

	return theme;
}

void
theme_destroy(Theme* theme)
{
	if(theme) 
	{	
		if(theme->name)
		{
			g_free(theme->name);
		}
		if(theme->path)
		{
			g_free(theme->path);
		}
		if(theme->obstacles) 
		{
			g_hash_table_foreach(theme->obstacles, 
					     (GHFunc) theme_destroy_image_table_item, NULL);
			g_hash_table_destroy(theme->obstacles);
		}
		if(theme->moveables)
		{
			g_hash_table_foreach(theme->moveables, 
					     (GHFunc) theme_destroy_image_table_item, NULL);
			g_hash_table_destroy(theme->moveables);
		}
		if(theme->connections)
		{
			g_hash_table_foreach(theme->connections, 
					     (GHFunc) theme_destroy_image_table_item, NULL);
			g_hash_table_destroy(theme->connections);
		}
		if(theme->decor)
		{
			g_hash_table_foreach(theme->decor, 
					     (GHFunc) theme_destroy_image_table_item, NULL);
			g_hash_table_destroy(theme->decor);
		}
		if(theme->bgcolor)
		{
			g_free(theme->bgcolor);	  
		}
		g_free(theme);
	}
}

void
theme_create_hash_table(void)
{
	gint i, n_dirs;
	gchar* users_home;
	struct dirent *dent = NULL;
	DIR* dir;
#ifdef DEBUG
	gchar* search_dirs[3];
#else
	gchar* search_dirs[2];
#endif

	/* init hash table */
	available_themes = g_hash_table_new((GHashFunc) g_str_hash, 
					    (GCompareFunc) g_str_equal);

	/* define directorys where to search for themes */
	users_home = g_get_home_dir();
	search_dirs[0] = g_strjoin("/", users_home, ".atomix/themes", NULL);
	search_dirs[1] = g_strjoin("/", DATADIR, "atomix/themes", NULL);

#ifdef DEBUG
	search_dirs[2] = g_strdup("../themes");
	n_dirs = 3;
#else
	n_dirs = 2;
#endif
    
	/* search the dirs */
	for(i = 0; i < n_dirs; i++)
	{
		dir = opendir(search_dirs[i]);
		if(dir)
		{
			char* filename;
			char* themename;
			char* directory;
			while((dent = readdir(dir)) != NULL)
			{
				if((g_strcasecmp(".", dent->d_name)!=0) &&
				   (g_strcasecmp("..", dent->d_name)!=0))
				{
					/* is current file a directory? */
					directory = g_concat_dir_and_file(search_dirs[i],
									  dent->d_name);
					if(g_file_test(directory, G_FILE_TEST_ISDIR))
					{
						/* try to load file */
						filename = g_concat_dir_and_file(directory, 
										 "theme");
						themename = theme_get_theme_name_xml(filename);
						theme_add_theme_name(themename, directory);
						g_free(filename);
						g_free(themename);
					}
					g_free(directory);
				}
			}
		}

		g_free(dent);
		closedir(dir);
	}

	for(i = 0; i < n_dirs; i++) g_free(search_dirs[i]);
}

void theme_add_theme_name(gchar *themename, gchar *directory)
{
	gchar *search_result;

	if((themename != NULL) && 
	   (directory != NULL))
	{
	
		/* don't add a theme twice */
		search_result = g_hash_table_lookup(available_themes, themename);
		
		if(search_result == NULL)
		{
			g_hash_table_insert(available_themes,
					    g_strdup(themename),
					    g_strdup(directory));

			g_print("found theme: %s\n", themename);
		}
	}
}

void
theme_destroy_hash_table(void)
{
	if(available_themes)
	{
		g_hash_table_foreach(available_themes, 
				     (GHFunc) theme_destroy_name_table_item, NULL);
		g_hash_table_destroy(available_themes);
	}
}

/*=================================================================
 
  Theme functions

  ---------------------------------------------------------------*/

GdkImlibImage*
theme_get_tile_image(Theme* theme, Tile *tile)
{
	gint type; 
	gint image_id;
	ThemeElement *element = NULL;

	g_return_val_if_fail(theme != NULL, NULL);
	g_return_val_if_fail(tile != NULL, NULL);

	type = tile_get_type(tile);
	image_id = tile_get_image_id(tile);

	switch(type)
	{
	case TILE_OBSTACLE:
		if(theme->obstacles!=NULL) 
		{
			element = g_hash_table_lookup(theme->obstacles, 
						      GINT_TO_POINTER(image_id));
		}
		break;
	    
	case TILE_MOVEABLE:
		if(theme->moveables)
		{
			element = g_hash_table_lookup(theme->moveables, 
						      GINT_TO_POINTER(image_id));
		}
		break;

	case TILE_DECOR:
		if(theme->decor)
		{
			element = g_hash_table_lookup(theme->decor, 
						      GINT_TO_POINTER(image_id));

		}
		break;
	    
	default:
	}

	return theme_get_element_image(theme, element);
}


GSList* theme_get_tile_connection_images(Theme *theme, Tile *tile)
{
	gint i;
	gint length;
	GSList *image_list = NULL;
	GdkImlibImage *image;
	ThemeElement *element;
	GSList *id_list;

	g_return_val_if_fail(theme!=NULL, NULL);
	g_return_val_if_fail(tile!=NULL, NULL);

	id_list = tile_get_connection_ids(tile);
	
	length = g_slist_length(id_list);
	for(i = 0; i  < length; i++)
	{
		/* get id of connection image */
		gint id = GPOINTER_TO_INT(g_slist_nth_data(id_list, i));

		/* get appropriate image name */
		element = g_hash_table_lookup(theme->connections, GINT_TO_POINTER(id));
		image = theme_get_element_image(theme, element);

		if(image)
		{
                       /* add image to list */
			image_list = g_slist_append(image_list, (gpointer) image);
		}
	}

	return image_list;
}

void
theme_get_tile_size(Theme *theme, gint *width, gint *height)
{
	if(theme)
	{
		*width = theme->tile_width;
		*height = theme->tile_height;
	}
	else
	{
		*width = 0;
		*height = 0;
	}
}

gchar* theme_get_background_color(Theme *theme)
{
	return theme->bgcolor;
}

/*=================================================================
 
  Theme load/save functions

  ---------------------------------------------------------------*/
Theme*
theme_load_xml(gchar *name/*, LoadResult *result*/)
{
	xmlDocPtr doc;
	xmlNodePtr node;
	Theme *theme;
	gchar *prop_value;
	gchar *dir_path;
	gchar *file_path;
	guint revision = 1;

	theme = theme_new();

	/* read file */
	dir_path = (gchar*) g_hash_table_lookup(available_themes, name);
	theme_set_path(theme, dir_path);
	file_path = g_strjoin("/", dir_path, "theme", NULL);
    
	doc = xmlParseFile(file_path); 
       
	if(doc != NULL)
	{
		node = doc->root;
		while(node!=NULL)
		{
			if(g_strcasecmp(node->name,"THEME")==0)
			{
				/* handle theme node */
				prop_value = xmlGetProp(node, "name");
				theme->name = g_strdup(prop_value);
				node = node->childs;
			}
			else 
			{	
				if(g_strcasecmp(node->name,"REVISION")==0)
				{
					/* handle revision node */
					gchar *content = xmlNodeGetContent(node);
					revision = atoi(content);
					g_free(content);
				}
				else if(g_strcasecmp(node->name,"MOVEABLE")==0)
				{
					/* handle moveable node */
					theme_fill_img_array(theme, node->childs, 
							     TILE_MOVEABLE,
							     revision);
				}
				else if(g_strcasecmp(node->name,"CONNECTION")==0)
				{
					/* handle connection node */
					theme_fill_connection_array(theme, node->childs, 
								    revision);
				}
				else if(g_strcasecmp(node->name,"OBSTACLE")==0)
				{
					/* handle obstacle node */
					theme_fill_img_array(theme, node->childs,
							     TILE_OBSTACLE,
							     revision);
				}
				else if(g_strcasecmp(node->name,"DECOR")==0)
				{
					/* handle decor node */
					theme_fill_img_array(theme, node->childs, 
							     TILE_DECOR,
							     revision);
				}
				else if(g_strcasecmp(node->name,"ANIMSTEP")==0)
				{
					/* handle animstep node */
					prop_value = xmlGetProp(node, "dist");
					theme->animstep = atoi(prop_value);
				}
				else if(g_strcasecmp(node->name,"BGCOLOR")==0)
				{
					/* handle background color */
					prop_value = xmlGetProp(node, "color");
					theme->bgcolor = g_strdup(prop_value);
				}
				else
				{
					g_print("theme.c: Unknown TAG, ignoring <%s>\n",
						node->name);
				}
		
				node = node->next;
			}
		}
	
		xmlFreeDoc(doc);
	}
	else
	{
		return NULL;
	}

	return theme;
}

void
theme_save_xml(Theme *theme, gchar *filename)
{
	xmlDocPtr doc;
	xmlAttrPtr attr;
	xmlNodePtr node;
	xmlNodePtr theme_node;
	xmlNodePtr obstacle_node;
	xmlNodePtr moveable_node;
	xmlNodePtr connection_node;
	/* xmlNodePtr decor_node;  not supported yet */
	gint dist;
	gchar *str_buffer;

	str_buffer = g_malloc(5*sizeof(gchar));
    
	/* create xml doc */    
	doc = xmlNewDoc("1.0");
	xmlSetDocCompressMode(doc, 9);

	/* theme name */
	theme_node = xmlNewDocNode(doc, NULL, "THEME", NULL);
	doc->root = theme_node;
	attr = xmlSetProp(theme_node, "name", g_strdup(theme->name));

	/* revision */
	node = xmlNewChild(theme_node, NULL, "REVISION", "2");

	/* bg color */
	node = xmlNewChild(theme_node, NULL, "BGCOLOR", NULL);
	attr = xmlSetProp(node, "color", g_strdup(theme->bgcolor));

	/* animstep */
	node = xmlNewChild(theme_node, NULL, "ANIMSTEP", NULL);
	dist = g_snprintf(str_buffer, 5, "%i", theme->animstep);
	attr = xmlSetProp(node, "dist", g_strdup(str_buffer));

	/* obstacle section */
	obstacle_node = xmlNewChild(theme_node, NULL, "OBSTACLE", NULL);
	g_hash_table_foreach(theme->obstacles, 
			     (GHFunc) theme_save_image_element_xml,
			     (gpointer) obstacle_node);

	/* moveable section */
	moveable_node = xmlNewChild(theme_node, NULL, "MOVEABLE", NULL);
	g_hash_table_foreach(theme->moveables, 
			     (GHFunc) theme_save_image_element_xml,
			     (gpointer) moveable_node);

	/* connection section */
	connection_node = xmlNewChild(theme_node, NULL, "CONNECTION", NULL);
	g_hash_table_foreach(theme->connections, 
			     (GHFunc) theme_save_image_element_xml,
			     (gpointer) connection_node);

	/* decor section */
	/* currently not used */
	// decor_node = xmlNewChild(theme_node, NULL, "DECOR", NULL);
	// for(i = 0; i < theme->decor->len; i++)
	// {
	// 	   child_node = xmlNewChild(decor_node, NULL, "IMAGE", NULL);
	// 	   length = g_snprintf(str_buffer, 5, "%i", i);
	// 	   attr = xmlSetProp(child_node, "no", g_strdup(str_buffer));
	// 
	// 	   img_src = (gchar*)g_ptr_array_index(theme->decor, i);
	// 	   attr = xmlSetProp(child_node, "src", g_strdup(img_src));
	// }

	xmlSaveFile(filename, doc);

	xmlFreeDoc(doc);
	g_free(str_buffer);
}

void theme_save_image_element_xml(gpointer key, gpointer value, gpointer user_data)
{
	gchar *str_buffer;
	gint length;
	gint id = GPOINTER_TO_INT(key);
	xmlNodePtr child;
	xmlNodePtr parent = (xmlNodePtr) user_data;
	xmlAttrPtr attr;
	ThemeElement *element = (ThemeElement*) value;
	
	str_buffer = g_malloc(5*sizeof(gchar));
	
	child = xmlNewChild(parent, NULL, "IMAGE", NULL);
	length = g_snprintf(str_buffer, 5, "%i", id);
	attr = xmlSetProp(child, "no", g_strdup(str_buffer));
	attr = xmlSetProp(child, "src", g_strdup(element->file));
}

gchar*
theme_get_theme_name_xml(gchar *filename)
{
	xmlDocPtr doc;
	xmlNodePtr node;
	gchar* name = NULL;
	gchar *prop_value;

	/* read file */
	doc = xmlParseFile(filename); 
       
	if(doc != NULL)
	{
		node = doc->root;
	
		if((node!=NULL) &&
		   (g_strcasecmp(node->name,"THEME")==0))
		{
			/* handle level node */
			prop_value = xmlGetProp(node, "name");
			name = g_strdup(prop_value);
		}
		
		xmlFreeDoc(doc);
	}

	return name;
}

/*=================================================================
 
  Internal helper functions

  ---------------------------------------------------------------*/
GdkImlibImage* theme_get_element_image(Theme *theme, ThemeElement *element)
{
	GdkImlibImage *image = NULL;

	g_return_val_if_fail(theme!=NULL, NULL);
	g_return_val_if_fail(element!=NULL, NULL);

	if(element->image == NULL)
	{
		if(!element->loading_failed)
		{
			/* try to load the image */
			gchar *full_path;

			/* add full path */
			full_path = g_strjoin("/", theme->path, element->file, NULL);
			element->image = gdk_imlib_load_image(full_path);

			if(element->image != NULL)
			{
				image = element->image;
			}
			else
			{
				/* loading failed */
				element->loading_failed = TRUE;
				image = NULL;
			}
		}
		else
		{
			/* loading failed last time */
			image = NULL;
		}
	}
	else
	{
		image = element->image;
	}
	
	return image;
}

void
theme_add_tile_image(Theme* theme, TileType type, gint id, ThemeElement *element)
{
	GHashTable *hash_table = NULL;
 
	g_return_if_fail(theme!=NULL);	
	g_return_if_fail(element!=NULL);

	/* get right hash table */
	switch(type)
	{
	case TILE_OBSTACLE:
		if(theme->obstacles) 
		{
			hash_table = theme->obstacles; 
		}
		break;
		
	case TILE_MOVEABLE:
		if(theme->moveables)
		{
			hash_table = theme->moveables;
		}
		break;

	case TILE_DECOR:
		if(theme->decor)
		{
			hash_table = theme->decor;
		}
		break;
	    
	default:
	}
		
	/* check whether the id already exists */
	if(g_hash_table_lookup(hash_table, GINT_TO_POINTER(id)) != NULL)
	{
		g_print("theme.c: Warning, duplicate id's ... ignoring.\n");
		theme_destroy_element(element);
		return;
	}

	/* add element with key id */
	g_hash_table_insert(hash_table, GINT_TO_POINTER(id),
			    (gpointer) element);


	/* obtain the width and height of the images */
	if((theme->tile_width == 0) || (theme->tile_height == 0))
	{
		gchar *full_path;

		full_path = g_strjoin("/", theme->path, element->file, NULL);
		element->image = gdk_imlib_load_image(full_path);
		if(element->image != NULL)
		{
			theme->tile_width = element->image->rgb_width;
			theme->tile_height = element->image->rgb_height;
		}
		else
		{
			element->loading_failed = TRUE;
		}

		g_free(full_path);
	}
}

void
theme_set_path(Theme* theme, gchar* path)
{
	theme->path = g_strdup(path);
}

void 
theme_fill_img_array(Theme *theme, xmlNodePtr node, TileType type, guint revision)
{
	gchar *img_src;
	gchar *id_str; 
	ThemeElement *element;
	gint id; 

	while(node != NULL)
	{
		/* get image number */
		id_str = xmlGetProp(node, "no");
		id = atoi(id_str);
		g_free(id_str);

		/* get image file */
		img_src = xmlGetProp(node, "src");

		/* create new theme element */
		element = g_malloc(sizeof(ThemeElement));
		element->id = id;
		element->file = img_src;
		element->name = "test";
		element->loading_failed = FALSE;
		element->image = NULL;

		theme_add_tile_image(theme, type, id, element);
		node = node->next;
	}
}

void
theme_add_connection_image(Theme* theme, gint id, ThemeElement *element)
{
	g_return_if_fail(theme != NULL);
	g_return_if_fail(element != NULL);

	/* check whether the id already exists */
	if(g_hash_table_lookup(theme->connections, GINT_TO_POINTER(id)) != NULL)
	{
		g_print("theme.c: Warning, duplicate id's ... ignoring.\n");
		theme_destroy_element(element);
		return;
	}

	/* add element with key id */
	g_hash_table_insert(theme->connections, GINT_TO_POINTER(id),
			    (gpointer) element);
}

void
theme_fill_connection_array(Theme *theme, xmlNodePtr node, guint revision)
{
	gchar *img_src;
	gchar *id_str; 
	ThemeElement *element;
	gint id; 

	while(node != NULL)
	{
		/* get image number */
		id_str = xmlGetProp(node, "no");
		id = atoi(id_str);
		g_free(id_str);

		/* get image file */
		img_src = xmlGetProp(node, "src");

		/* create new theme element */
		element = g_malloc(sizeof(ThemeElement));
		element->id = id;
		element->file = img_src;
		element->name = "test";
		element->loading_failed = FALSE;
		element->image = NULL;

		theme_add_connection_image(theme, id, element);
		node = node->next;
	}
}


void
theme_print_hash_table(gpointer key, gpointer value, gpointer user_data)
{
	g_print("Key: %s, Value: %s\n",(gchar*) key, (gchar*) value);
}



void
theme_destroy_name_table_item(gpointer key, gpointer value, gpointer user_data)
{
	g_free(key);
	g_free(value);
}

void
theme_destroy_image_table_item(gpointer key, gpointer value, gpointer user_data)
{
	theme_destroy_element((ThemeElement*) value);
}


void
theme_destroy_element(ThemeElement *element)
{
	g_return_if_fail(element != NULL);

	if(element->file)
	{
		g_free(element->file);
	}
	if(element->image)
	{
		gdk_imlib_destroy_image(element->image);
	}
	if(element->name)
	{
		// currently the name is static
		// g_free(element->name);
	}
	g_free(element);
}


